/** 
 * \file Edt.cpp
 *  
 * \brief  Implementation of  a  class for  representing  a 2D  binary
 * image.
 *
 * \version 1.0
 * \date January 2016
 *
 * \copyright (C) 2016 Orthrous Group.
 *
 * \attention This  software is provided 'as-is',  without any express
 *            or  implied warranty. In  no event  will the  authors be
 *            held liable for any damages arising from the use of this
 *            software.
 *
 *            Permission is granted to anyone to use this software for
 *            any purpose,  including commercial applications,  and to
 *            alter  it and  redistribute  it freely,  subject to  the
 *            following restrictions:
 *            
 *            1.   The   origin  of   this   software   must  not   be
 *            misrepresented; you  must not  claim that you  wrote the
 *            original  software.  If  you  use  this  software  in  a
 *            product, an acknowledgment  in the product documentation
 *            would be appreciated but is not required.
 *
 *            2.  Altered  source versions  must be plainly  marked as
 *            such,  and  must  not  be misrepresented  as  being  the
 *            original software.
 *
 *            3. This  notice may not  be removed or altered  from any
 *            source distribution.
 */

#include "Edt.h"            // Edt

#include <cassert>          // assert
#include <algorithm>        // std::max
#include <cstddef>          // size_t
#include <vector>           // std::vector
#include <cmath>            // sqrt


/** 
 * \defgroup EDTNameSpace Namespace edt.
 * @{
 */

/**
 * \namespace edt
 *
 * \brief Namespace edt contains  the definition and implementation of
 * classes to compute  the exact Euclidean distance transform  of a 2D
 * binary image.
 */

namespace edt {

  /**
   * \fn Edt::Edt( const Image& image )
   *
   * \brief Creates an instance of this class.
   *
   * \param image A 2D binary image.
   *
   */
  Edt::Edt( const Image& image )
  {
    /*
     * Set the number of columns and rows.
     */
    _cols = image.get_number_of_columns() ;
    _rows = image.get_number_of_rows() ;
    
    /*
     * Set origin coordinates.
     */
    _xorg = image.get_origin_x_coordinate() ;
    _yorg = image.get_origin_y_coordinate() ;
    
    /*
     * Set spacing.
     */
    _spac = image.get_spacing_between_rows_and_columns() ;
    
    /*
     * Allocate space for storing the distance values.
     */
    _fval = ( double** ) new double*[ _cols ] ;
    
    for( size_t i = 0 ; i < _cols ; i++ ) {
      _fval[ i ] = ( double* ) new double[ _rows ] ;
      for ( size_t j = 0 ; j < _rows ; j++ ) {
        _fval[ i ][ j ] = 0 ;
      }
    }

    /*
     * Compute the distance transform of the given image.
     */
    compute( image ) ;
    
    return ;
  }
     
    
  /**
   * \fn Edt::Edt( const Edt& dt )
   *
   * \brief Clones an instance of this class.
   *
   * \param dt A reference to another instance of this class.
   *
   */
  Edt::Edt( const Edt& dt )
  {
    /*
     * Set the number of columns and rows.
     */
    _cols = dt.get_number_of_columns() ;
    _rows = dt.get_number_of_rows() ;
    
    /*
     * Set origin coordinates.
     */
    _xorg = dt.get_origin_x_coordinate() ;
    _yorg = dt.get_origin_y_coordinate() ;
    
    /*
     * Set spacing.
     */
    _spac = dt.get_spacing_between_rows_and_columns() ;
    
    /*
     * Allocate space for storing the distance values.
     */
    _fval = ( double** ) new double*[ _cols ] ;
    
    for( size_t i = 0 ; i < _cols ; i++ ) {
      _fval[ i ] = ( double* ) new double[ _rows ] ;
      for ( size_t j = 0 ; j < _rows ; j++ ) {
        _fval[ i ][ j ] = dt.get_distance( i , j ) ;
      }
    }
    
    return ;
  }


  /**
   * \fn virtual Edt::~Edt()
   *
   * \brief Releases the memory held by an instance of this class.
   */
  Edt::~Edt()
  {
    if ( _fval != 0 ) {
      for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
        if ( _fval[ i ] != 0 ) {
          delete[] _fval[ i ] ;
        }
      }
      
      delete[] _fval ;
    }
    
    return ;
  }


  /**
   * \fn void Edt::compute( const Image& image )
   *
   * \brief Computes the exact euclidean distance from each pixel of a
   * given  2D binary  image to  the closest  point in  the continuous
   * analog of the digital boundary between foreground and background.
   *
   * \param image A 2D binary image.
   *
   */
  void
  Edt::compute( const Image& image )
  {
    /*
     * Get the maximum between the number of rows and columns.
     */
    size_t infinity = std::max(
                               image.get_number_of_columns() ,
                               image.get_number_of_rows()
                              ) ;
    
    /*
     * Create a temporary grid.
     */
    double** grid = ( double** ) new double*[ get_number_of_columns() ] ;
    
    infinity = 2 * ( infinity * infinity ) ;
    
    for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      grid[ i ] = ( double* ) new double[ get_number_of_rows() ] ;
      for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
        if ( image.is_pixel_in_foreground( i , j ) ) {
          grid[ i ][ j ] = 0 ;
        }
        else {
          grid[ i ][ j ] = infinity ;
        }
      }
    }
    
    /*
     * Compute the  distance from  background pixels to  their closest
     * foreground pixels.
     */
    
    for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
      voronoi(
              COLUMN ,
              j ,
              infinity ,
              grid
             ) ;
    }

    for ( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      voronoi(
              ROW ,
              i ,
              infinity ,
              grid
             ) ;
    }
    
    /*
     * Copy  the  distance values  of  the  background pixels  to  the
     * distance grid.
     */
    
    for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
        if ( !image.is_pixel_in_foreground( i , j ) ) {
          set_distance(
                       i ,
                       j ,
                       get_spacing_between_rows_and_columns() * sqrt( grid[ i ][ j ] )
                      ) ;
        }
      }
    }
    
    /*
     * Compute the  distance from  foreground pixels to  their closest
     * background pixels.
     */
    
    for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
        if ( image.is_pixel_in_foreground( i , j ) ) {
          grid[ i ][ j ] = infinity ;
        }
        else {
          grid[ i ][ j ] = 0 ;
        }
      }
    }

    for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
      voronoi(
              COLUMN ,
              j ,
              infinity ,
              grid
             ) ;
    }
    
    for ( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      voronoi(
              ROW ,
              i ,
              infinity ,
              grid
             ) ;
    }
    
    /*
     * Copy  the  distance values  of  the  foreground pixels  to  the
     * distance grid.
     */
    
    for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      for ( size_t j = 0 ; j < get_number_of_rows() ; j++ ) {
        if ( image.is_pixel_in_foreground( i , j ) ) {
          set_distance(
                       i ,
                       j ,
                       -get_spacing_between_rows_and_columns() * sqrt( grid[ i ][ j ] )
                      ) ;
        }
      }
    }

    /*
     * Release memory associated with the temporary grid.
     */

    for( size_t i = 0 ; i < get_number_of_columns() ; i++ ) {
      delete[] grid[ i ] ;
    }
      
    delete[] grid ;

    return ;
  }


  /**
   * \fn void Edt::voronoi( DIRECTION d , size_t r , double infty , double** grid )
   *
   * \brief  Builds a  partial  Voronoi diagram  whose  sites are  the
   * boundary pixels  of the  digital boundary between  foreground and
   * background of a given 2D binary image. This diagram speeds up the
   * computation of the squared Euclidean distance from an image pixel
   * to its closest point in the continuous analog of the boundary.
   *
   * \param d A column or row selector.
   * \param r A column or row index.
   * \param infty The number of pixels in the selected row or column.
   * \param grid A grid to store distance values.
   *
   */
  void
  Edt::voronoi(
               DIRECTION d ,
               size_t r ,
               double infty ,
               double** grid
              )
  {
    size_t nd ;
    
    if ( d == COLUMN ) {
      nd = get_number_of_columns() ;
    }
    else {
#ifdef DEBUGMODE
      assert( d == ROW ) ;
#endif
      nd = get_number_of_rows() ;
    }
    
    std::vector< double > g( nd , 0 ) ;
    std::vector< size_t > h( nd , 0 ) ;
    
    int l = -1 ;
    
    for ( size_t k = 0 ; k < nd ; k++ ) {
      double di = ( d == COLUMN )
                ? grid[ k ][ r ]
                : grid[ r ][ k ] ;
      
      if ( di != infty ) {
        while (
               ( l >= 1 )
               &&
               remove(
                      g[ l - 1 ] ,
                      g[ l ] ,
                      di ,
                      h[ l - 1 ] ,
                      h[ l ] ,
                      k
                     )
             )
        {
          --l;
        }
        
        ++l ;
        g[ size_t( l ) ] = di ;
        h[ size_t( l ) ] =  k ;
      }
    }
    
    if ( l == -1 ) return ;
    
    size_t ns = size_t( l ) ;
    size_t j = 0 ;
    
    for ( size_t k = 0 ; k < nd ; k++ ) {
      double h1 = double( h[ j ] ) - double( k ) ;
      double d1 = g[ j ] + ( h1 * h1 ) ;

      if ( j < ns ) {
        double h2 = double( h[ j + 1 ] ) - double( k ) ;
        double d2 = g[ j + 1 ] + ( h2 * h2 ) ;
        
        while ( ( j < ns ) && ( d1 > d2 ) ) {
          ++j ;
          d1 = d2 ;
          h2 = double( h[ j + 1 ] ) - double( k ) ;
          d2 = g[ j + 1 ] +  ( h2 * h2 ) ;
        }
      }
      
      if ( d == COLUMN ) {
        grid[ k ][ r ] = d1 ;
      }
      else {
        grid[ r ][ k ] = d1 ;
      }
    }
    
    return ;
  }

}

/** @} */ //end of group class.
