/** 
 * \file Edt.h
 *  
 * \brief Definition of a class for representing a 2D binary image.
 *
 * \version 1.0
 * \date January 2016
 *
 * \copyright (C) 2016 Orthrous Group.
 *
 * \attention This  software is provided 'as-is',  without any express
 *            or  implied warranty. In  no event  will the  authors be
 *            held liable for any damages arising from the use of this
 *            software.
 *
 *            Permission is granted to anyone to use this software for
 *            any purpose,  including commercial applications,  and to
 *            alter  it and  redistribute  it freely,  subject to  the
 *            following restrictions:
 *            
 *            1.   The   origin  of   this   software   must  not   be
 *            misrepresented; you  must not  claim that you  wrote the
 *            original  software.  If  you  use  this  software  in  a
 *            product, an acknowledgment  in the product documentation
 *            would be appreciated but is not required.
 *
 *            2.  Altered  source versions  must be plainly  marked as
 *            such,  and  must  not  be misrepresented  as  being  the
 *            original software.
 *
 *            3. This  notice may not  be removed or altered  from any
 *            source distribution.
 */


#ifndef __Edt_h
#define __Edt_h

#include "Image.h"          // Image

#include <cassert>          // assert
#include <cmath>            // fabs
#include <cstddef>          // size_t



/** 
 * \defgroup EDTNameSpace Namespace edt.
 * @{
 */

/**
 * \namespace edt
 *
 * \brief Namespace edt contains  the definition and implementation of
 * classes to compute  the exact Euclidean distance transform  of a 2D
 * binary image.
 */

namespace edt {

  /**
   * \class Edt
   *
   * \brief This class computes the exact Euclidean distance transform
   * of a 2D  binary image.  The class is based  on an algorithm given
   * in
   *
   * Maurer, C. R.; Qi, R.; Raghavan, V.
   *
   * A Linear  Time Algorithm  for Computing Exact  Euclidean Distance
   * Transforms of  Binary Images in Arbitrary  Dimensions, IEEE PAMI,
   * 25(2), 2003.
   *
   */

  class Edt {
  public:
    
    // -----------------------------------------------------------------
    //
    // Type definitions
    //
    // -----------------------------------------------------------------

    /**
     * \typedef DIRECTION
     *
     * \brief Type definition for table traversal direction.
     */
    typedef enum { COLUMN , ROW } DIRECTION ;
    
  protected:

    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------

    size_t    _cols ;  ///< Number of columns of the image.
    size_t    _rows ;  ///< Number of rows of the image.    
    double    _xorg ;  ///< X coordinate of the lower-left image pixel.
    double    _yorg ;  ///< Y coordinate of the lower-left image pixel.
    double    _spac ;  ///< Distance between two consecutive image pixels. 
    double**  _fval ;  ///< A two-dimensional array to store the distance values associated with each pixel.


  public:

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    /**
     * \fn Edt( const Image& image )
     *
     * \brief Creates an instance of this class.
     *
     * \param image A 2D binary image.
     *
     */
    Edt( const Image& image ) ;
     
    
    /**
     * \fn Edt( const Edt& dt )
     *
     * \brief Clones an instance of this class.
     *
     * \param dt A reference to another instance of this class.
     *
     */
    Edt( const Edt& dt ) ;


    /**
     * \fn virtual ~Edt()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~Edt() ;
    
    
    /**
     * \fn size_t get_number_of_columns() const
     *
     * \brief Returns the number of columns of the distance grid.
     *
     * \return The number of columns of the distance grid.
     *
     */
    size_t get_number_of_columns() const
    {
      return _cols ;
    }
    
    
    /**
     * \fn size_t get_number_of_rows() const
     *
     * \brief Returns the number of rows of the distance grid.
     *
     * \return The number of rows of the distance grid.
     *
     */
    size_t get_number_of_rows() const
    {
      return _rows ;
    }
    
    
    /**
     * \fn double get_origin_x_coordinate() const
     *
     * \brief Returns the X coordinate of the lower left distance grid
     * point.
     *
     * \return The X coordinate of the lower left distance grid point.
     *
     */
    double get_origin_x_coordinate() const
    {
      return _xorg ;
    }
    
    
    /**
     * \fn double get_origin_y_coordinate() const
     *
     * \brief Returns the Y coordinate of the lower left distance grid
     * point.
     *
     * \return The Y coordinate of the lower left distance grid point.
     *
     */
    double get_origin_y_coordinate() const
    {
      return _yorg ;
    }
    
    
    /**
     * \fn double get_spacing_between_rows_and_columns() const
     *
     * \brief Returns the distance between two consecutive pixels.
     *
     * \return The distance between two consecutive pixels.
     *
     */
    double  get_spacing_between_rows_and_columns() const
    {
      return _spac ;
    }
    
    
    /**
     * \fn double get_distance( size_t i , size_t j ) const
     *
     * \brief Returns the distance associated with the point at column
     * \c i and row \c j of the distance grid.
     *
     * \return The distance  associated with the point at  column \c i
     * and row \c j of the distance grid.
     *
     */
    double get_distance(
                        size_t i ,
                        size_t j
                       )
      const
    {
#ifdef DEBUGMODE
      assert( i < _cols ) ;
      assert( j < _rows ) ;
#endif
      
      return _fval[ i ][ j ] ;
    }


  protected:

    // ---------------------------------------------------------------
    //
    // Protected methods
    //
    // ---------------------------------------------------------------
    
    /**
     * \fn void set_distance( size_t i , size_t j , double value ) const
     *
     * \brief Assigns a  distance to the point at row  \c i and column
     * \c j of the distance grid.
     *
     * \param i The column of a pixel.
     * \param j The row of a pixel.
     * \param value The distance to be assigned to the pixel at column
     * i and row j.
     *
     */
    void set_distance(
                      size_t i ,
                      size_t j ,
                      double value
                     )
      const
    {
#ifdef DEBUGMODE
      assert( i < _cols ) ;
      assert( j < _rows ) ;
#endif 

      _fval[ i ][ j ] = value ;
      
      return ;
    }


    /**
     * \fn void compute( const Image& image )
     *
     * \brief Computes the exact euclidean distance from each pixel of
     * a given 2D binary image to  the closest point in the continuous
     * analog  of   the  digital   boundary  between   foreground  and
     * background.
     *
     * \param image A 2D binary image.
     *
     */
    void compute( const Image& image ) ;


    /**
     * \fn void voronoi( DIRECTION d , size_t r , double infty , double** grid )
     *
     * \brief Builds  a partial  Voronoi diagram  whose sites  are the
     * boundary pixels of the  digital boundary between foreground and
     * background of a  given 2D binary image. This  diagram speeds up
     * the computation of the squared Euclidean distance from an image
     * pixel  to its  closest point  in the  continuous analog  of the
     * boundary.
     *
     * \param d A column or row selector.
     * \param r A column or row index.
     * \param infty The number of pixels in the selected row or column.
     * \param grid A grid to store distance values.
     *
     */
    void voronoi(
                 DIRECTION d ,
                 size_t r ,
                 double infty ,
                 double** grid
                ) ;
  
 
    /**
     * \fn bool remove( double d1 , double d2 , double df , size_t x1 , size_t x2 , size_t xf ) const
     *
     * \brief Returns  the logic value true  if the Voronoi cell  of a
     * digital boundary pixel does not intersect a given row or column
     * of the image. Otherwise, it returns the logic value false.
     *
     * \param d1 
     * \param d2
     * \param df
     * \param x1
     * \param x2 
     * \param xf 
     *
     * \return The logic  value true if the Voronoi cell  of a digital
     * boundary pixel does not intersect a  given row or column of the
     * image. Otherwise, the logic value false.
     *
     */
    bool remove(
                double d1 ,
                double d2 ,
                double df ,
                size_t x1 ,
                size_t x2 ,
                size_t xf
               )
    const
    {
      const double a = double( x2 - x1 ) ;
      const double b = double( xf - x2 ) ;
      const double c = double( xf - x1 ) ;
      
      double res = ( c * fabs( d2 ) )
                 - ( b * fabs( d1 ) )
                 - ( a * fabs( df ) )
                 - (      a * b * c ) ;
      
      return res > 0 ;
    }

  } ;

}  

/** @} */ //end of group class.

#endif   // __Edt_h
