/** 
 * \file Image.cpp
 *  
 * \brief  Implementation of  a  class for  representing  a 2D  binary
 * image.
 *
 * \version 1.0
 * \date January 2016
 *
 * \copyright (C) 2016 Orthrous Group.
 *
 * \attention This  software is provided 'as-is',  without any express
 *            or  implied warranty. In  no event  will the  authors be
 *            held liable for any damages arising from the use of this
 *            software.
 *
 *            Permission is granted to anyone to use this software for
 *            any purpose,  including commercial applications,  and to
 *            alter  it and  redistribute  it freely,  subject to  the
 *            following restrictions:
 *            
 *            1.   The   origin  of   this   software   must  not   be
 *            misrepresented; you  must not  claim that you  wrote the
 *            original  software.  If  you  use  this  software  in  a
 *            product, an acknowledgment  in the product documentation
 *            would be appreciated but is not required.
 *
 *            2.  Altered  source versions  must be plainly  marked as
 *            such,  and  must  not  be misrepresented  as  being  the
 *            original software.
 *
 *            3. This  notice may not  be removed or altered  from any
 *            source distribution.
 */

#include "Image.h"          // Image

#include <cassert>          // assert
#include <cstddef>          // size_t


/** 
 * \defgroup EDTNameSpace Namespace edt.
 * @{
 */

/**
 * \namespace edt
 *
 * \brief Namespace edt contains  the definition and implementation of
 * classes to compute  the exact Euclidean distance transform  of a 2D
 * binary image.
 */

namespace edt {


  /**
   * \fn Image::Image( size_t sx , size_t sy , double xo , double yo , double sp )
   *
   * \brief Creates an instance of this class.
   *
   * \param sx The number of columns of the image.
   * \param sy The number of rows of the image.
   * \param xo X coordinate of the lower-left image pixel.
   * \param yo Y coordinate of the lower-left image pixel.
   * \param sp Distance between two consecutive image pixels.
   *
   */
  Image::Image(
               size_t sx ,
               size_t sy ,
               double xo ,
               double yo ,
               double sp
              )
  {
    /*
     * Set the number of columns and rows.
     */
    _cols = sx ;
    _rows = sy ;
    
    /*
     * Set origin coordinates.
     */
    _xorg = xo ;
    _yorg = yo ;
    
    /*
     * Set spacing.
     */
    _spac = sp ;
    
    
    /*
     * Allocate space for storing the pixel values.
     */
    _fval = ( char** ) new char*[ _cols ] ;
    
    for( size_t i = 0 ; i < _cols ; i++ ) {
      _fval[ i ] = ( char* ) new char[ _rows ] ;
      for ( size_t j = 0 ; j < _rows ; j++ ) {
        _fval[ i ][ j ] = '0' ;
      }
    }
    
    return ;
  }
  
  
  /**
   * \fn Image::Image( const Image& image )
   *
   * \brief Clones an instance of this class.
   *
   * \param image A reference to another instance of this class.
   *
   */
  Image::Image( const Image& image )
  {
    /*
     * Set dimension, origin coordinates and spacing.
     */
    _cols = image._cols ;
    _rows = image._rows ;
    _xorg = image._xorg ;
    _yorg = image._yorg ;
    _spac = image._spac ;
    
    /*
     * Allocate space for storing the pixel values.
     */
    _fval = ( char** ) new char*[ _cols ] ;
    
    for( size_t i = 0 ; i < _cols ; i++ ) {
      _fval[ i ] = ( char* ) new char[ _rows ] ;
      for ( size_t j = 0 ; j < _rows ; j++ ) {
        _fval[ i ][ j ] = image._fval[ i ][ j ] ;
      }
    }
    
    return ;
  }
  
  
  /**
   * \fn virtual Image::~Image()
   *
   * \brief Releases the memory held by an instance of this class.
   */
  Image::~Image()
  {
    if ( _fval != 0 ) {
      for( size_t i = 0 ; i < _cols ; i++ ) {
        if ( _fval[ i ] != 0 ) {
          delete[] _fval[ i ] ;
        }
      }
      
      delete[] _fval ;
    }
    
    return ;
  }
  
  
  /**
   * \fn bool Image::is_pixel_on_digital_boundary( size_t i , size_t j ) const
   *
   * \brief Returns the  logic value true if the image  pixel at row i
   * and  column j  belongs  to  the digital  boundary  of the  image;
   * otherwise, returns the logic value false.
   *
   * \param i The column of a pixel.
   * \param j The row of a pixel.
   *
   * \return The  logic value  true if  the image pixel  at row  i and
   * column j belongs to the digital boundary of the image; otherwise,
   * the logic value false.
   *
   */
  bool
  Image::is_pixel_on_digital_boundary(
                                      size_t i ,
                                      size_t j
                                     )
    const
  {
#ifdef DEBUGMODE
    assert( i < _cols ) ;
    assert( j < _rows ) ;
#endif
    
    bool p0 = is_pixel_in_foreground(
                                     i ,
                                     j
                                    ) ;
    
    if ( i > 0 ) {
      bool p1 = is_pixel_in_foreground(
                                       i - 1 ,
                                       j
                                      ) ;
      if ( p0 != p1 ) {
        return true ;
      }
    }
    
    if ( i < ( _cols - 1 ) ) {
      bool p1 = is_pixel_in_foreground(
                                       i + 1 ,
                                       j
                                      ) ;
      if ( p0 != p1 ) {
        return true ;
      }
    }
    
    if ( j > 0 ) {
      bool p1 = is_pixel_in_foreground(
                                       i ,
                                       j - 1
                                      ) ;
      if ( p0 != p1 ) {
        return true ;
      }
    }
    
    if ( j < ( _rows - 1 ) ) {
      bool p1 = is_pixel_in_foreground(
                                       i ,
                                       j + 1
                                      ) ;
      if ( p0 != p1 ) {
        return true ;
      }
    }
    
    return false;    
  }
  

}

/** @} */ //end of group class.

