/** 
 * \file Image.h
 *  
 * \brief Definition of a class for representing a 2D binary image.
 *
 * \version 1.0
 * \date January 2016
 *
 * \copyright (C) 2016 Orthrous Group.
 *
 * \attention This  software is provided 'as-is',  without any express
 *            or  implied warranty. In  no event  will the  authors be
 *            held liable for any damages arising from the use of this
 *            software.
 *
 *            Permission is granted to anyone to use this software for
 *            any purpose,  including commercial applications,  and to
 *            alter  it and  redistribute  it freely,  subject to  the
 *            following restrictions:
 *            
 *            1.   The   origin  of   this   software   must  not   be
 *            misrepresented; you  must not  claim that you  wrote the
 *            original  software.  If  you  use  this  software  in  a
 *            product, an acknowledgment  in the product documentation
 *            would be appreciated but is not required.
 *
 *            2.  Altered  source versions  must be plainly  marked as
 *            such,  and  must  not  be misrepresented  as  being  the
 *            original software.
 *
 *            3. This  notice may not  be removed or altered  from any
 *            source distribution.
 */


#ifndef __Image_h
#define __Image_h


#include <cassert>          // assert
#include <cstddef>          // size_t


/** 
 * \defgroup EDTNameSpace Namespace edt.
 * @{
 */

/**
 * \namespace edt
 *
 * \brief Namespace edt contains  the definition and implementation of
 * classes to compute  the exact Euclidean distance transform  of a 2D
 * binary image.
 */

namespace edt {

  /**
   * \class Image
   *
   * \brief This  class represents a  2D binary image.  The continuous
   * analog of the  image pixels are assumed to be  squares (i.e., the
   * image is assumed to be isotropic).
   */

  class Image {
  protected:

    // ---------------------------------------------------------------
    //
    // Protected data members
    //
    // ---------------------------------------------------------------

    size_t  _cols ;  ///< Number of columns of the image.
    size_t  _rows ;  ///< Number of rows of the image.    
    double  _xorg ;  ///< X coordinate of the lower-left image pixel.
    double  _yorg ;  ///< Y coordinate of the lower-left image pixel.
    double  _spac ;  ///< Distance between two consecutive image pixels. 
    char**  _fval ;  ///< A two-dimensional array to store the values associated with each pixel.


  public:

    // ---------------------------------------------------------------
    //
    // Public methods
    //
    // ---------------------------------------------------------------

    
    /**
     * \fn Image( size_t sx , size_t sy , double xo , double yo , double sp )
     *
     * \brief Creates an instance of this class.
     *
     * \param sx The number of columns of the image.
     * \param sy The number of rows of the image.
     * \param xo X coordinate of the lower-left image pixel.
     * \param yo Y coordinate of the lower-left image pixel.
     * \param sp Distance between two consecutive image pixels.
     *
     */
    Image(
          size_t sx ,
          size_t sy ,
          double xo ,
          double yo ,
          double sp
         ) ;
    
    
    /**
     * \fn Image( const Image& image )
     *
     * \brief Clones an instance of this class.
     *
     * \param image A reference to another instance of this class.
     *
     */
    Image( const Image& image ) ;
    
    
    /**
     * \fn virtual ~Image()
     *
     * \brief Releases the memory held by an instance of this class.
     */
    virtual ~Image() ;
    
    
    /**
     * \fn size_t get_number_of_columns() const
     *
     * \brief Returns the number of columns of this image.
     *
     * \return The number of columns of this image.
     *
     */
    size_t get_number_of_columns() const
    {
      return _cols ;
    }
    
    
    /**
     * \fn size_t get_number_of_rows() const
     *
     * \brief Returns the number of rows of this image.
     *
     * \return The number of rows of this image.
     *
     */
    size_t get_number_of_rows() const
    {
      return _rows ;
    }
    
    
    /**
     * \fn double get_origin_x_coordinate() const
     *
     * \brief Returns the X coordinate of the lower left image pixel.
     *
     * \return The X coordinate of the lower left image pixel.
     *
     */
    double get_origin_x_coordinate() const
    {
      return _xorg ;
    }
    
    
    /**
     * \fn double get_origin_y_coordinate() const
     *
     * \brief Returns the Y coordinate of the lower left image pixel.
     *
     * \return The Y coordinate of the lower left image pixel.
     *
     */
    double get_origin_y_coordinate() const
    {
      return _yorg ;
    }
    
    
    /**
     * \fn double get_spacing_between_rows_and_columns() const
     *
     * \brief Returns the distance between two consecutive pixels.
     *
     * \return The distance between two consecutive pixels.
     *
     */
    double  get_spacing_between_rows_and_columns() const
    {
      return _spac ;
    }
    
    
    /**
     * \fn char get_value( size_t i , size_t j ) const
     *
     * \brief Returns the value (i.e., either the character '0' or the
     * character '1') associated with the pixel at  row i and column j
     * of the image.
     *
     * \return  The  value (i.e.,  either  the  character '0'  or  the
     * character '1') associated with the pixel  at row i and column j
     * of the grid.
     *
     */
    char get_value(
                   size_t i ,
                   size_t j
                  )
    const
    {
#ifdef DEBUGMODE
      assert( i < _cols ) ;
      assert( j < _rows ) ;
#endif
      
      return _fval[ i ][ j ] ;
    }
    
    
    /**
     * \fn void set_value( size_t i , size_t j , char value )
     *
     * \brief Assigns a  value (i.e., either the character  '0' or the
     * character  '1') to  the pixel  at row  i and  column j  of this
     * image.
     *
     * \param i The column of a pixel.
     * \param j The row of a pixel.
     * \param value The value to be  assigned to the pixel at column i
     * and row j.
     *
     */
    void set_value(
                   size_t i ,
                   size_t j ,
                   char value
                  )
    {
#ifdef DEBUGMODE
      assert( i < _cols ) ;
      assert( j < _rows ) ;
      assert( ( value == '0' ) || ( value == '1' ) ) ;
#endif
      
      _fval[ i ][ j ] = value ;
      
      return ;
    }
    
    
    /**
     * \fn bool is_pixel_on_digital_boundary( size_t i , size_t j ) const
     *
     * \brief Returns the logic value true if the image pixel at row i
     * and  column j  belongs to  the digital  boundary of  the image;
     * otherwise, returns the logic value false.
     *
     * \param i The column of a pixel.
     * \param j The row of a pixel.
     *
     * \return The  logic value true if  the image pixel at  row i and
     * column  j  belongs  to  the  digital  boundary  of  the  image;
     * otherwise, the logic value false.
     *
     */
    bool is_pixel_on_digital_boundary(
                                      size_t i ,
                                      size_t j
                                     )
      const ;
    
    
    /**
     * \fn bool is_pixel_in_foreground( size_t i , size_t j ) const
     *
     * \brief Returns the logic value true if the image pixel at row i
     * and column j belongs to the foreground of the image; otherwise,
     * returns the logic value false.
     *
     * \param i The column of a pixel.
     * \param j The row of a pixel.
     *
     * \return The  logic value true if  the image pixel at  row i and
     * column j belongs to the foreground of the image; otherwise, the
     * logic value false.
     */
    bool is_pixel_in_foreground(
                                size_t i ,
                                size_t j
                               )
      const
    {
#ifdef DEBUGMODE
      assert( i < _cols ) ;
      assert( j < _rows ) ;
#endif
      
      return ( _fval[ i ][ j ] == '1' ) ;
    }
    
    
    /**
     * \fn bool is_pixel_in_background( size_t i , size_t j ) const
     *
     * \brief Returns the logic value true if the image pixel at row i
     * and column j belongs to the background of the image; otherwise,
     * returns the logic value false.
     *
     * \param i The column of a pixel.
     * \param j The row of a pixel.
     *
     * \return The  logic value true if  the image pixel at  row i and
     * column j belongs to the background of the image; otherwise, the
     * logic value false.
     */
    bool is_pixel_in_background(
                                size_t i ,
                                size_t j
                               )
    const
    {
      return !is_pixel_in_foreground( i , j ) ;
    }

  } ;
    
}  

/** @} */ //end of group class.

#endif   // __Image_h
