/** 
 * \file main.cpp
 *  
 * \brief A program to compute  the exact Euclidean distance transform
 * of a given  2D binary image. The image is  assumed to be isotropic,
 * i.e., the  distance between consecutive  pixels in the same  row or
 * column is constant.
 *
 * \version 1.0
 * \date January 2016
 *
 * \copyright (C) 2016 Orthrous Group.
 *
 * \attention This  software is provided 'as-is',  without any express
 *            or  implied warranty. In  no event  will the  authors be
 *            held liable for any damages arising from the use of this
 *            software.
 *
 *            Permission is granted to anyone to use this software for
 *            any purpose,  including commercial applications,  and to
 *            alter  it and  redistribute  it freely,  subject to  the
 *            following restrictions:
 *            
 *            1.   The   origin  of   this   software   must  not   be
 *            misrepresented; you  must not  claim that you  wrote the
 *            original  software.  If  you  use  this  software  in  a
 *            product, an acknowledgment  in the product documentation
 *            would be appreciated but is not required.
 *
 *            2.  Altered  source versions  must be plainly  marked as
 *            such,  and  must  not  be misrepresented  as  being  the
 *            original software.
 *
 *            3. This  notice may not  be removed or altered  from any
 *            source distribution.
 */

#include <iostream>                 // std::cout, std::endl, std::cerr
#include <fstream>                  // std::ifstream, std::ofstream
#include <string>                   // std::string
#include <cstdlib>                  // exit, atoi, EXIT_SUCCESS, EXIT_FAILURE
#include <iomanip>                  // std::setprecision
#include <cassert>                  // assert
#include <ctime>                    // time, clock, CLOCKS_PER_SEC, clock_t
#include <cstddef>                  // size_t

#include "Image.h"                  // edt::Image
#include "Edt.h"                    // edt::Edt


/**
 * \fn void read_input( const std::string& filename , edt::Image*& image )
 *
 * \brief Read in a 2D digital binary image and create a grid. 
 *
 * \param filename The name of a file describing a 2D binary image.
 * \param image A reference to an image object.
 *
 */
void read_input(
                const std::string& filename ,
                edt::Image*& image
               ) ;


/**
 * \fn void write_edt( const std::string& filename , const edt::Edt& dt )
 *
 * \brief Write the  distance values obtained by  a distance transform
 * applied to a 2D binary image to an output file.
 *
 * \param filename The name of the output file.
 * \param dt A grid representing distance values.
 *
 */
void write_edt(
               const std::string& filename ,
               const edt::Edt& dt
              ) ;


/**
 * \fn void write_image( const std::string& filename , const edt::Edt& dt )
 *
 * \brief Create a visual  representation of a distance transformation
 * applies to a  2D binary image.  The representation is  written to a
 * VTK file.
 *
 * \param  filename The  name of the output file.
 * \param dt A grid representing distance values.
 *
 */
void write_image(
                 const std::string& filename ,
                 const edt::Edt& dt
                ) ;


/**
 * \fn int main( int argc , char* argv[]  )
 *
 * \brief Computes the  exact Euclidean distance transform  of a given
 * 2D binary image.
 *
 * \param argc The number of command-line arguments.
 * \param argv An array with the command-line arguments.
 *
 * \return An integer number.
 */
int main( int argc , char* argv[] ) {
  //
  // Check command-line arguments.
  //

  using std::cin  ;
  using std::cout ;
  using std::cerr ;
  using std::endl ;
  
  if ( argc != 4 ) {
    cerr << "Usage: "
         << endl
         << "\t\t edt arg1 arg2 arg3"
         << endl
         << "\t\t arg1: name of a file describing a binary image"
         << endl
         << "\t\t arg2: name of the output file with the distance values at the image pixels"
         << endl
	 << "\t\t arg3: name of a VTK file describing a visual representation of the distance transform image"
         << endl
         << endl ;
    cerr.flush() ;
    
    return EXIT_FAILURE ;
  }

  //
  // Read in the input file.
  //
  
  clock_t start, end ;

  cerr << endl
       << "Reading file describing a 2D binary image..." 
       << endl ;
  cerr.flush() ;

  std::string fn1( argv[ 1 ] ) ;
  
  edt::Image* image ;

  start = clock() ;
  read_input( fn1 , image ) ;
  end = clock() ;

  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl 
       << endl ;
  cerr.flush() ;
  
  //
  // Compute the exact Euclidean distance transform.
  //

  cerr << "Compute the exact Euclidean distance transform... " 
       << endl ;
  cerr.flush() ;

  start = clock() ;
  
  edt::Edt edt( *image ) ;

  end = clock() ;

  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl 
       << endl ;
  cerr.flush() ;

  //
  // Release memory
  //

  cerr << "Releasing memory..." 
       << endl 
       << endl ;
  cerr.flush() ;

  if ( image != 0 ) delete image ;

  //
  // Write the distance transform information to an output file.
  //

  cerr << "Writing out the distance transform information to an output file..." 
       << endl ;
  cerr.flush() ;

  start = clock() ;

  std::string fn2( argv[ 2 ] ) ;

  write_edt(
	    fn2 ,
	    edt
           ) ;

  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl 
       << endl ;
  cerr.flush() ;

  //
  // Write a visual representation of the distance transform to an output file.
  //

  cerr << "Writing out a visual representation of the distance transform to an output file..." 
       << endl ;
  cerr.flush() ;

  start = clock() ;

  std::string fn3( argv[ 3 ] ) ;

  write_image(
              fn3 ,
              edt
             ) ;

  cerr << ( (double) ( end - start ) ) / CLOCKS_PER_SEC 
       << " seconds." 
       << endl 
       << endl ;
  cerr.flush() ;

  //
  // Done.
  //

  cerr << "Finished."
       << endl 
       << endl
       << endl ;
  cerr.flush() ;
  
  return EXIT_SUCCESS ;
}


/**
 * \fn void read_input( const std::string& filename , edt::Image*& image )
 *
 * \brief Read in a 2D digital binary image and create a grid. 
 *
 * \param filename The name of a file describing a 2D binary image.
 * \param image A reference to an image object.
 *
 */
void read_input(
                const std::string& filename ,
                edt::Image*& image
               )
{
  //
  // Open the input file
  //
  std::ifstream in( filename.c_str() ) ;
  
  if ( in.is_open() ) {
    //
    // Read in the number of cols and rows of the image.
    //
    size_t cols ;
    size_t rows ;

    in >> cols >> rows ;

    //
    // Read in the coordinates of the lower left pixel of the image.
    //
    double xo ;
    double yo ;

    in >> xo >> yo ;

    //
    // Read in the distances between consecutive pixels.
    //
    double sx ;
    double sy ;
    
    in >> sx >> sy ;

    assert( sx == sy ) ;

    //
    // Read in the image content.
    //
    
    image = new edt::Image(
                           cols ,
                           rows ,
                           xo ,
                           yo ,
                           sx 
                          ) ;

    for ( size_t j = 0 ; j < rows ; j++ ) {
      for ( size_t i = 0 ; i < cols ; i++ ) {
        char ch ;
        in >> ch ;
        image->set_value(
                         i ,
                         j ,
                         ch
                        ) ;
      }
    }

    //
    // Close file
    //
    
    in.close() ;
  }
  
  return ;  
}


/**
 * \fn void write_edt( const std::string& filename , const edt::Edt& dt )
 *
 * \brief Write the  distance values obtained by  a distance transform
 * applied to a 2D binary image to an output file.
 *
 * \param filename The name of the output file.
 * \param dt A grid representing distance values.
 *
 */
void write_edt(
               const std::string& filename ,
               const edt::Edt& dt
              )
{
  std::ofstream ou( filename.c_str() ) ;
  
  if ( ou.is_open() ) {
    //
    // Get grid information.
    //
    size_t cols = dt.get_number_of_columns() ;
    size_t rows = dt.get_number_of_rows() ;
    
    //
    // Print out the point and cell information to a VTK file.
    //
    
    ou << std::setprecision( 18 ) << std::fixed ;
    
    //
    // Print out the file header.
    //
    
    ou << cols
       << '\t'
       << rows
       << std::endl ;

    ou << dt.get_origin_x_coordinate()
       << '\t'
       << dt.get_origin_y_coordinate() 
       << std::endl ;

    ou << dt.get_spacing_between_rows_and_columns()
       << '\t'
       << dt.get_spacing_between_rows_and_columns()
       << std::endl ;
   
    //
    // Print out the distance values per row.
    //
    
    for ( size_t j = 0; j < rows ; j++ ) {
      for ( size_t i = 0; i < cols ; i++ ) {
        ou << dt.get_distance( i , j )
           << '\t' ;
      }

      ou << std::endl ;
    }
    
    //
    // Close file
    //
    
    ou.close() ;
  }
  
  return ;
}


/**
 * \fn void write_image( const std::string& filename , const edt::Edt& dt )
 *
 * \brief Create a visual  representation of a distance transformation
 * applies to a  2D binary image.  The representation is  written to a
 * VTK file.
 *
 * \param  filename The  name of the output file.
 * \param dt A grid representing distance values.
 *
 */
void write_image(
		 const std::string& filename ,
		 const edt::Edt& dt
		)
{
  std::ofstream ou( filename.c_str() ) ;
  
  if ( ou.is_open() ) {
    //
    // Get grid information.
    //
    size_t cols = dt.get_number_of_columns() ;
    size_t rows = dt.get_number_of_rows() ;
    
    double xo = dt.get_origin_x_coordinate() ;
    double yo = dt.get_origin_y_coordinate() ;

    double sp = dt.get_spacing_between_rows_and_columns() ;
    
    //
    // Print out the point and cell information to a VTK file.
    //
    
    ou << std::setprecision( 18 ) << std::fixed ;
    
    //
    // Print out the file header.
    //
    
    ou << "# vtk DataFile Version 1.0" << std::endl ;
    ou << "this is a file generated by the application edt" << std::endl ;
    ou << "ASCII" << std::endl ;
    ou << "DATASET UNSTRUCTURED_GRID" << std::endl ;
    
    ou << std::endl ;
    
    //
    // Print out the number of vertices.
    //

    size_t np = ( cols + 1 ) * ( rows + 1 ) ;
    
    ou << "POINTS "
       << np
       << " FLOAT"
       << std::endl ;
    
    //
    // Print out the vertex coordinates.
    //
    
    for ( size_t j = 0; j <= rows ; j++ ) {
      double y = yo + ( j * sp ) - ( sp * 0.5 ) ;
      for ( size_t i = 0; i <= cols ; i++ ) {
        double x = xo + ( i * sp ) - ( sp * 0.5 ) ;
        ou << x
           << '\t'
           << y
           << '\t'
           << 0
           << std::endl ;
      }
    }
    
    ou << std::endl ;
    
    //
    // Print out the number of cells.
    //

    size_t nf = cols * rows ;
    
    ou << "CELLS "
       << nf
       << " "
       << 5 * nf 
       << std::endl ;
    
    //
    // Print out the cell information.
    //

    for ( size_t j = 0; j < rows ; j++ ) {
      size_t k = j * ( cols + 1 ) ;
      for ( size_t i = 0; i < cols ; i++ ) {
        size_t h = i + k ;
        ou << "4 "
           << h
           << '\t'
           << h + 1
           << '\t'
           << h + ( cols + 2 )
           << '\t'
           << h + ( cols + 1 )
           << std::endl ;
      }
    }
    
    ou << std::endl ;
    
    ou << "CELL_TYPES "
       << nf
       << std::endl ;
    
    for ( unsigned i = 0; i < nf ; i++ ) {
      ou << "9 " ;
    }
    
    ou << std::endl << std::endl ;
    
    //
    // Print out color information
    //
    
    ou << "CELL_DATA "
       << nf
       << std::endl ;
    
    ou << "SCALARS colors int 1"
       << std::endl ;
    
    ou << "LOOKUP_TABLE default"
       << std::endl ;

    for ( size_t j = 0 ; j < rows ; j++ ) {
      for ( size_t i = 0 ; i < cols ; i++ ) {
        double val = dt.get_distance( i , j ) ;

#ifdef DEBUGMODE
        assert( val != 0 ) ;
#endif
        
        if ( val < 0 ) {
          ou << "2" << std::endl ;
        }
        else {
          ou << "1" << std::endl ;
        }
      }
    }
    
    //
    // Close file
    //
    
    ou.close() ;
  }
  
  return ;
}
